(* 
 * Copyright (c) 2001, 2002 The University of Wroclaw.
 * All rights reserved. See file COPYRIGHT for details about licensing.
 *
 * $Id: gontcodegen.ml,v 1.23 2002/05/27 15:08:33 malekith Exp $
 *
 * Module implementation: Gontcodegen
 * Description: This module generally converts Gontutil.* -> Ksi.*, i.e. 
 * generates code. Key things to remember is that some variables can
 * be in closures, and it is sometimes necessary to convert names between
 * Gont and Ksi namespace.
 *
 * All artificial identifiers created should contain '$' symbol in it.
 * You cannot start symbols with '$' though. Also don't include '$$'
 * in names, as it is reserved for modules.
 *)

open Gonttree
open Gontutil
open Gontbind
open Gonttypes
open Gontmodule

type fnc_ctx = {ctx_level: int;
                ctx_closure_type: id;
                ctx_aux_decls: Ksi.decl list;
                ctx_fnname: id;
                ctx_module: id}
                

type ctx = {gctx_globals: symbol list;
            gctx_global_types: typedef list;
            gctx_module: id;
            gctx_file: string;
            gctx_init_fnc: string;
            gctx_fini_fnc: string}

let global_ctx = {ctx_level = 0; 
                  ctx_closure_type = "";
                  ctx_aux_decls = [];
                  ctx_fnname = "???";
                  ctx_module = "???"}
;;

(* This is here mainly to remind you, that you're changing namespace.
 * You go from Gont to Ksi names, which in most cases are the same,
 * but they surly doesn't have to. That's why Ksi.id type is hidden. 
 *)
let ksi_name s = 
  let e = String.copy s in
  for i = 0 to String.length e - 1 do
    if e.[i] = ':' then e.[i] <- '$'
  done;
  Ksi.make_id e
(*
 tmp = exp;

 int x = 3;
 switch (f(g(h()))) {
 case F[1]: print
 case F[2]:
 case F[x]: err; x = -1;
 }
 x = 3;
 exp = tmp;
*)
let lambda_exp = Ksi.Struct(ksi_name "gont_lambda_exp")
let ksi_closure = Ksi.Ref(ksi_name "cl$")
let ksi_closure_var n = Ksi.Arrow(ksi_closure, ksi_name n)
let ksi_parent_closure = Ksi.Ref(ksi_name "parent_cl$")
;;

let created_structs : (field_record list, id) Hashtbl.t = Hashtbl.create 16
let used_exns : symbol list ref = ref []
;;

let init_codegen m f =
  Hashtbl.clear created_structs;
  used_exns := [];
  let (s, t) = Gontmodule.globals m in
  {gctx_module = m;
   gctx_globals = s;
   gctx_global_types = t;
   gctx_file = f;
   gctx_init_fnc = "";
   gctx_fini_fnc = ""}
;;

let get_module_info c =
  {mi_name = c.gctx_module;
   mi_used_exns = (List.map (fun s -> s.sym_name) !used_exns);
   mi_used_modules = Gontbind.used_modules_list ();
   mi_init_fnc = 
     if c.gctx_init_fnc = "" then 
       "" 
     else 
       c.gctx_module ^ "$$" ^ c.gctx_init_fnc;
   mi_fini_fnc =
     if c.gctx_fini_fnc = "" then 
       "" 
     else 
       c.gctx_module ^ "$$" ^ c.gctx_fini_fnc;
  }
  
let global_check loc ctx sym =
  let rec check l =
    match l with
        ps::xs when ps.sym_name = sym.sym_name ->
           (try
              global_subtype ps.sym_type sym.sym_type
            with Type_error e -> 
               err (loc, xf "type of %s is incompatible with type \
                             declared in interface: %s" 
                             sym.sym_name e));
           xs
      | x::xs ->
             x :: (check xs)
      | [] -> [] in
   let new_g = check ctx.gctx_globals in
     if List.length new_g = List.length ctx.gctx_globals then
       (ctx, [Ksi.Local]) 
     else
       ({ctx with gctx_globals = new_g}, [Ksi.Public])

let global_type_check loc ctx td =
  let rec check l =
    match l with
        ({td_name = n} as p)::xs when n = td.td_name ->
           (try
              match p.td_type with
                  AbstractType -> ()
                | _ -> global_subtype p.td_type td.td_type
            with Type_error e -> 
               err (loc, xf "type of %s is incompatible with type \
                             declared in interface: %s" 
                             td.td_name e));
           xs
      | x::xs ->
             x :: (check xs)
      | [] -> [] in
   let new_g = check ctx.gctx_global_types in
     if List.length new_g = List.length ctx.gctx_globals then
       (ctx, [Ksi.Local])
     else
       ({ctx with gctx_global_types = new_g}, [Ksi.Public])
;;

let const_int i =
        Ksi.IntConst(Ksi.ptroff (), string_of_int(i))

let const_uint i =
        Ksi.IntConst(Ksi.uptroff (), string_of_int(i))

let ksi_null = Ksi.Null(Ksi.Ptr(Ksi.Void))


let ksi_runtime_proto {gctx_module = m; gctx_file=f} =
  let fnc (r, n, a) = Ksi.FuncProto([Ksi.Extern], r, n, a) in
  let noret_fnc (r, n, a) = 
        Ksi.FuncProto([Ksi.Extern; Ksi.NoReturn], r, n, a) in
  let ptr = Ksi.Ptr(Ksi.Void) in
  let size_t = Ksi.uptroff () in
  let gstring = Ksi.Ptr(Ksi.Struct(ksi_name "gont_string")) in
[
  Ksi.TopComment("implementation of module " ^ m);
  Ksi.InputFile(f);
  Ksi.TopComment("BEG Runtime proto");
  Ksi.StructDef(ksi_name "gont_lambda_exp", [], 
                [(ptr, ksi_name "fnc"); 
                 (ptr, ksi_name "closure")]);
  fnc(ptr, ksi_name "gont_xmalloc", [size_t]);
  fnc(ptr, ksi_name "gont_xmalloc0", [size_t]);
  fnc(Ksi.Ptr(lambda_exp), ksi_name "gont_make_lambda", [ptr; ptr]);
  Ksi.StructDef(ksi_name "gont_string", [], 
                [(size_t, ksi_name "len"); 
                 (Ksi.Ptr(Ksi.UInt(8)), ksi_name "data")]);
  fnc(gstring, ksi_name "gont_make_string", [size_t; Ksi.Ptr(Ksi.UInt(8))]);
  fnc(gstring, ksi_name "gont_strjoin", [gstring; gstring]);
  fnc(Ksi.ptroff (), ksi_name "gont_strcmp", [gstring; gstring]);
  fnc(Ksi.ptroff (), ksi_name "gont_streq", [gstring; gstring]);
  (* We lay, that it returns pointer, but it doesn't matter, since it
   * doesn't return :) *)
  noret_fnc(ptr, ksi_name "gont_null_access", []);
  fnc(Ksi.Ptr(Ksi.c_type Ksi.C_double), 
      ksi_name "gont_make_float",
      [Ksi.c_type Ksi.C_double]);
  fnc(Ksi.Ptr(Ksi.Ptr(Ksi.Void)), 
      ksi_name "gont_make_tuple",
      [Ksi.uptroff(); Ksi.Ellipsis]);
  Ksi.StructDef(ksi_name "gont_union", [], 
                [(Ksi.uptroff(), ksi_name "field_no"); 
                 (Ksi.ptroff(), ksi_name "value")]);
  fnc(Ksi.Ptr(Ksi.Struct(ksi_name "gont_union")), 
      ksi_name "gont_make_union",
      [Ksi.uptroff(); ptr]);
  fnc(Ksi.Ptr(Ksi.Void), ksi_name "gont_push_try_block", []);
  fnc(Ksi.Ptr(Ksi.Void), ksi_name "gont_pop_raised_value", []);
  fnc(Ksi.Void, ksi_name "gont_pop_try_block", []);
  noret_fnc(Ksi.Void, ksi_name "gont_raise", [Ksi.Ptr(Ksi.Void)]);
  noret_fnc(Ksi.Void, ksi_name "gont_match_failure", []);
  Ksi.Global([Ksi.Extern], Ksi.ptroff(), 
             ksi_name (m ^ "$$__go_file_linked__$"));
  Ksi.Func([Ksi.Public], Ksi.Void, ksi_name (m ^ "$$__dummy__$"), [],
           Ksi.Bind(Ksi.Modify(Ksi.Ref(ksi_name (m ^ "$$__go_file_linked__$")),
                               const_int 1)));
  Ksi.TopComment("END Runtime proto");
]
;;

let aux_decls : Ksi.decl list ref = ref []

(* This function converts all tyvars to 'a (because all tyvars are the
 * same at the Ksi level... and we want exactly the same ksi structure 
 * to be generated for each. *)
let drop_tyvars t =
  let f t =
    match t with
    | TyVar _ -> TyVar "a"
    | t -> t in
  type_walk f t
;;

let extern_name sym =
  let n = 
    try
      match (List.find (function ExternName _ -> true | _ -> false)   
                       sym.sym_properties) with
      | ExternName x -> x
      | _ -> assert false
    with Not_found -> sym.sym_name in
  ksi_name n

(* As we use structural equality for structures, we need to generate
 * structure definition (in C's sense) each time we use structure. However
 * we cache definitions here. *)
let rec ksi_struct l =
  try
    Hashtbl.find created_structs l
  with Not_found ->
    let seq = (seq_name "struct") ^ "$" in
    (* Add struct to hash first, in case it is referenced from itself. *)
    Hashtbl.add created_structs l seq;
    let struct_decl =
        Ksi.StructDef(ksi_name seq, [], 
                      List.map (fun f -> 
                                 match f with 
                                 | {fr_type = t; fr_sym = Some sym} -> 
                                        (ksi_local_type t, 
                                         extern_name sym)
                                 | _ -> assert false)
                               l) in
    aux_decls := struct_decl :: !aux_decls; 
    seq

(* Return Ksi type, to be used as type of local variable.
 * This is sometimes not exactly, what is passed to and returned from
 * functions, when called. *)
and ksi_local_type t =
  let t = drop_tyvars t in
  match t with
    | Int -> Ksi.ptroff ()
    | Bool -> Ksi.Bool
    | AbstractType -> Ksi.Ptr(Ksi.Void)
       (* raise (Error "cannot have variable of abstract type") *)
    | Void -> 
       raise (Error "cannot have variable of void type")
    | Float -> Ksi.c_type(Ksi.C_double)
    | Func _ -> Ksi.Ptr(Ksi.Struct(ksi_name "gont_lambda_exp"))
    | String -> Ksi.Ptr(Ksi.Struct(ksi_name "gont_string"))
    | TyVar(_) -> Ksi.Ptr(Ksi.Void)
     (* raise (Error "cannot have variable of unspecified type") *)
    | Tuple(_) -> Ksi.Ptr(Ksi.Ptr(Ksi.Void))
    | Struct l
    | OptStruct l -> Ksi.Ptr(Ksi.Struct(ksi_name (ksi_struct l)))
    | Exn
    | Union _ -> Ksi.Ptr(Ksi.Struct(ksi_name "gont_union"))
    | NamedType _ -> ksi_local_type (instance_value t)

(* Return Ksi type, to be used when passing arguments, returning values. *)
and ksi_passed_type t =
  let t = drop_tyvars t in
  match t with
    | Int -> Ksi.ptroff ()
    | Bool -> Ksi.ptroff ()
    | Void -> Ksi.Void
    | Float -> Ksi.Ptr(Ksi.c_type(Ksi.C_double))
    | Func _ -> Ksi.Ptr(Ksi.Struct(ksi_name "gont_lambda_exp"))
    | String -> Ksi.Ptr(Ksi.Struct(ksi_name "gont_string"))
    | TyVar(_) -> Ksi.Ptr(Ksi.Void)
    | Tuple(_) -> Ksi.Ptr(Ksi.Ptr(Ksi.Void))
    | Struct l
    | OptStruct l -> Ksi.Ptr(Ksi.Struct(ksi_name (ksi_struct l)))
    | Exn
    | Union _ -> Ksi.Ptr(Ksi.Struct(ksi_name "gont_union"))
    | NamedType _ -> ksi_passed_type (instance_value t)
    | AbstractType -> Ksi.Ptr(Ksi.Void)

let ksi_noptr_type t =
  let t = drop_tyvars t in
  match t with
  | Struct(l) -> Ksi.Struct(ksi_name (ksi_struct l))
  | Union(l) -> Ksi.Struct(ksi_name "gont_union")
  | OptStruct(l) -> Ksi.Struct(ksi_name (ksi_struct l))
  | _ -> ksi_local_type t
;;

(* Append prototype of function (or declaration of external variable)
 * to aux_decls. Only do so once, for given symbol.
 *)
let push_proto (sym:symbol) =
  if has_property sym IsPrototyped then ()
  else if has_property sym IsException then 
    begin
      let decl = Ksi.Global([Ksi.Extern], 
                            Ksi.uptroff (), 
                            extern_name sym) in
      aux_decls := decl :: !aux_decls;
      used_exns := sym :: !used_exns;
      set_property sym IsPrototyped
    end
  else if (not (has_property sym IsNonLocal)) then ()
  else begin
    let decl =
      match instance_value sym.sym_type with
      | Func {ft_ret_type=r; ft_parms=a} ->
          Ksi.FuncProto([Ksi.Extern], 
                        ksi_passed_type r,
                        extern_name sym,
                        (List.map ksi_passed_type a))
      | t ->
          Ksi.Global([Ksi.Extern], 
                     ksi_passed_type t, 
                     extern_name sym) in
    aux_decls := decl :: !aux_decls;
    set_property sym IsPrototyped
  end
;;

let ksi_op2 o =
  match o with
    | Plus -> Ksi.Plus
    | Minus -> Ksi.Minus
    | Mult -> Ksi.Mult
    | Div -> Ksi.TruncDiv
    | Mod -> Ksi.TruncMod
    | Lt -> Ksi.Lt
    | Gt -> Ksi.Gt
    | Le -> Ksi.Le
    | Ge -> Ksi.Ge
    | Eq -> Ksi.Eq
    | Ne -> Ksi.Ne
    | Assign -> assert false
  
let ksi_op1 o e =
  match o with
  | Negate -> Ksi.Unary(Ksi.Negate, e)
  | PreDec -> Ksi.Binary(Ksi.PreDec, e, const_int 1)
  | PreInc -> Ksi.Binary(Ksi.PreInc, e, const_int 1)
  | PostDec -> Ksi.Binary(Ksi.PostDec, e, const_int 1)
  | PostInc -> Ksi.Binary(Ksi.PostInc, e, const_int 1)
;;

let rec ksi_stmt (c, s) =
  let (c', s') =
    match s.raw_stmt with
        Skip ->
          (c, Ksi.Nop)
      | Exp(e) ->
          let (c', e') = ksi_exp (c, e) in
          (* Type of statement has to be void. *)
          (c', Ksi.Convert(Ksi.Void, e'))
      | Return(e) ->
          let (c', e') = ksi_exp (c, e) in
          let e'' = 
            match type_of e with
                Float -> Ksi.Call(ksi_name "gont_make_float", [e'])
              | Bool -> Ksi.Convert(ksi_passed_type Bool, e')
              | _ -> e' in
          (c', Ksi.Return(e''))
      | ReturnVoid ->
          (c, Ksi.Return(Ksi.Nop))
      | Block([]) ->
          (c, Ksi.Nop)
      | Block([s]) ->
          let (c', s') = ksi_stmt (c, s) in (c', Ksi.Bind(s'))
      | Block([s1; s2]) ->
          let (c1, s1') = ksi_stmt (c, s1) in
          let (c2, s2') = ksi_stmt (c1, s2) in
          (c2, Ksi.Bind(Ksi.Compound(s1', s2')))
      | Block(s) ->
          let (c', s') = list_apply ksi_stmt (c, s) in
          (c', Ksi.Bind(Ksi.CompoundList(s')))
      | VarDef v ->
          (match v.var_symbol with
              Some sym -> 
                let ts = "type = " ^ (string_of_type sym.sym_type) ^ 
                         ", instance = " ^ 
                         (string_of_type (instance_value sym.sym_type)) in
                if has_property sym IsInClosure then
                  (c, Ksi.Comment(sym.sym_name ^ " in closure " ^ ts, Ksi.Nop))
                else
                  (c, Ksi.Comment(ts, 
                                  Ksi.Var(ksi_local_type sym.sym_type, 
                                          extern_name sym)))
            | None -> assert false)
      | Pair (a, b) ->
          let (c', a') = ksi_stmt (c, a) in
          let (c'', b') = ksi_stmt (c', b) in
          (c'', Ksi.Compound(a', b'))
      | If(e, s1, s2) ->
          let (c1, e') = ksi_exp (c, e) in
          let (c2, s1') = ksi_stmt (c1, s1) in
          let (c3, s2') = ksi_stmt (c2, s2) in
          (c3, Ksi.Cond(Ksi.Convert(Ksi.Bool, e'), s1', s2'))
      | BreakIfNot(e) ->
          let (c', e') = ksi_exp (c, e) in
          (c', Ksi.ExitLoop(Ksi.Unary(Ksi.TruthNot, e')))
      | Loop {ll_sym = Some sym; ll_body = body; ll_inc = inc} ->
          let (c, body') = ksi_stmt (c, body) in
          let (c, inc') = ksi_stmt (c, inc) in
          let b = Ksi.Block(ksi_name (sym.sym_name ^ "$continue"), body') in
          let b = match inc' with
                  | Ksi.Nop -> b
                  | _ -> Ksi.Compound(b, inc') in
          let code = Ksi.Block(ksi_name (sym.sym_name ^ "$break"), 
                               Ksi.Loop(b)) in
          (c, code)
      | Break {br_sym = Some sym; br_continue = cn} ->
          let n = sym.sym_name ^ "$" ^ (if cn then "continue" else "break") in
          (c, Ksi.Exit(ksi_name n))
      | Break {br_sym = None}
      | Loop {ll_sym = None} -> assert false
      | Switch (ex, casel) ->
          let ex_var = seq_name "switch_ex$" in
          let ex_ptr = Ksi.Ref(ksi_name ex_var) in
          let (c, exp) = ksi_exp (c, ex) in
          let (c, s) = pattern_matching (c, false, ex_ptr, casel) in
          (c, 
            Ksi.Bind (Ksi.CompoundList [
                Ksi.Var (ksi_local_type (type_of ex), ksi_name ex_var);
                Ksi.Modify (ex_ptr, exp); s]))
      | Let (patel, st) ->
          let (c, assigned_vars) = 
            match_let_patterns(c, patel, 0) in 
          let (c, sta) = ksi_stmt (c, st) in
          (c, Ksi.Bind (Ksi.CompoundList (assigned_vars @ [sta])))
      | Raise e ->
          let (c, e') = ksi_exp (c, e) in
          (c, Ksi.Call (ksi_name "gont_raise", [e']))
      | Try t ->
          let exc = ksi_name (seq_name "exn$") in
          let try_finally e f = 
            let try_block = Ksi.Call(ksi_name "gont_push_try_block", []) in
            let cond = Ksi.Binary(Ksi.Eq, 
                                  Ksi.BuiltIn(ksi_name "setjmp", [try_block]), 
                                  const_int 0) in
            let exc_val = Ksi.Call(ksi_name "gont_pop_raised_value", []) in
            let assign = Ksi.Modify(Ksi.Ref(exc), exc_val) in
            let pop_block = Ksi.Call(ksi_name "gont_pop_try_block", []) in
            let ec = Ksi.Compound(e, pop_block) in
            let ex = Ksi.Cond(cond, ec, Ksi.Convert(Ksi.Void, assign)) in
            Ksi.TryFinally(ex, f) in
          let if_e th =
            Ksi.Cond(Ksi.Binary(Ksi.Ne, Ksi.Ref(exc), ksi_null),
                     th,
                     Ksi.Nop) in
          (* FIXME: optimize if finally = Skip, or with_matching = [] *)
          let (c, body) = ksi_stmt (c, t.try_body) in
          let (c, with_matching) = 
                pattern_matching (c, true, Ksi.Ref(exc), t.try_with) in
          let with_matching = if_e with_matching in
          let (c, finally) = ksi_stmt (c, t.try_finally) in
          let re_raise = 
                if_e (Ksi.Call(ksi_name "gont_raise", [Ksi.Ref(exc)])) in
          let code = 
              try_finally 
                body 
                (try_finally 
                  with_matching 
                  (try_finally 
                    finally
                    re_raise)) in
          let var_def = Ksi.Var(ksi_local_type Exn, exc) in
          let var_init = Ksi.Convert(Ksi.Void, 
                                     Ksi.Modify(Ksi.Ref(exc), ksi_null)) in
          let all = Ksi.Bind(Ksi.CompoundList [var_def; var_init; code]) in
          (c, all)
      | Open _ -> (c, Ksi.Nop) in
  (c', Ksi.Line(s.stmt_loc.loc_line, s'))

(* Return *value* of given variable *)
and find_in_pattern var_name pat ksi_ptr =
    match pat with
    | VoidPattern -> raise Not_found
    | VarPattern {vp_name = n} ->
      if n = var_name then
        ksi_ptr
      else
        raise Not_found
    | ExprPattern _ -> raise Not_found
    | TuplePattern {tpr_patterns = pats} ->
      let rec test_one idx lst =
        match lst with
        | {raw_pattern = x}::xs ->
          (try
            find_in_pattern var_name x
                (* *( ((void ** )ksi_ptr) + (idx * sizeof(void* )) ) *)
                (Ksi.Indirect(Ksi.Binary(Ksi.Plus, 
                        Ksi.Convert(Ksi.Ptr(Ksi.Ptr(Ksi.Void)), ksi_ptr), 
                        Ksi.Binary(Ksi.Mult, 
                                   const_uint idx, 
                                   Ksi.SizeOf(Ksi.Ptr(Ksi.Void))))))
          with Not_found -> test_one (idx+1) xs)
        | [] -> raise Not_found in
        test_one 0 pats
    | UnionPattern un ->
      find_in_pattern var_name un.upr_pattern.raw_pattern
        (* (((struct un* )ksi_ptr)->value) *)
           (Ksi.Arrow(
                Ksi.Convert(Ksi.Ptr(Ksi.Struct(ksi_name "gont_union")), 
                            ksi_ptr),
                ksi_name "value"))
    
and pattern_matching (c, exn_matching, ex_ptr, cali) =
  match cali with
    {cc_pattern = pat; cc_stmt = st}::xs ->
    let allocate_patt_var sym =
      let ksi_type_of_var = Ksi.Ptr(ksi_local_type sym.sym_type) in
      let value = Ksi.Convert(ksi_type_of_var, 
                            Ksi.Unary(Ksi.Addr, 
                            find_in_pattern sym.sym_name
                               pat.raw_pattern 
                               ex_ptr)) in
      Ksi.Compound(Ksi.Var(ksi_type_of_var, extern_name sym),
                   Ksi.Modify(Ksi.Ref(extern_name sym), value)) in
      (match pat.raw_pattern with
        VarPattern {vp_symbol = Some sym} ->
          set_property sym IsTopPattern
      | _ -> ());
      let varlist = strip_patt_var pat in
      let (c', condi) = match_it (c, ex_ptr, pat) in 
      let matched_vars =
        Ksi.CompoundList(List.map allocate_patt_var varlist) in
      let (c'', stm1) = ksi_stmt (c', st) in
      let clr_exn = 
        if exn_matching then
          Ksi.Convert(Ksi.Void, Ksi.Modify(ex_ptr, ksi_null))
        else
          Ksi.Nop in
      let stm3 = Ksi.Bind(Ksi.CompoundList [matched_vars; stm1; clr_exn]) in
      let (c3, stm2) = pattern_matching(c'', exn_matching, ex_ptr, xs) in
      (c3, Ksi.Cond(condi, stm3, stm2))
  | [] -> 
      if exn_matching then 
        (c, Ksi.Nop)
      else
        (c, Ksi.Call(ksi_name "gont_match_failure", []))

and match_it (c, exp_ptr, pat) =
  match pat.raw_pattern with
    | VoidPattern -> (c, Ksi.BoolConst(true))
    | VarPattern _ ->
        (c, Ksi.BoolConst(true))
    | ExprPattern ex -> 
        let (c', ex') = ksi_exp (c, ex) in
        let converted_ex = Ksi.Convert(ksi_local_type (type_of ex), exp_ptr) in
        let code =
          match instance_value (type_of ex) with
          | String ->
             Ksi.Convert(Ksi.Bool, Ksi.Call(ksi_name "gont_streq",
                                            [converted_ex;
                                             ex']))
          | Int | Float | Bool -> 
             Ksi.Binary(Ksi.Eq, converted_ex, ex') 
          | _ -> assert false in
        (c', code)
    | TuplePattern {tpr_patterns = pats} ->
      let rec match_one (c, idx, lst) =
        match lst with
        | x::xs ->
            let part_ptr = (Ksi.Indirect(Ksi.Binary(Ksi.Plus, 
                            Ksi.Convert(Ksi.Ptr(Ksi.Ptr(Ksi.Void)), exp_ptr), 
                            Ksi.Binary(Ksi.Mult, 
                                       const_uint idx, 
                                       Ksi.SizeOf(Ksi.Ptr(Ksi.Void)))))) in
            let (c', cond1) = match_it (c, part_ptr, x) in
            let (c'', cond2) = match_one(c', idx+1, xs) in
            (c', Ksi.Binary(Ksi.TruthAndIf, cond1, cond2))
        | [] -> (c,  Ksi.BoolConst(true)) in
        match_one (c, 0, pats)
    | UnionPattern {upr_sym = Some sym; upr_pattern = val_pat} ->
        let un_ptr = Ksi.Convert(Ksi.Ptr(Ksi.Struct(ksi_name "gont_union")), 
                                 exp_ptr) in
        let ksi_val_ptr = 
        (* (((struct un* )ksi_ptr)->value) *)
            Ksi.Arrow(un_ptr, ksi_name "value") in
        let ksi_f_id = 
        (* ((struct un* )ksi_ptr)->field_no *)
          Ksi.Arrow(un_ptr, ksi_name "field_no") in
        let ksi_un_id =
          if has_property sym IsException then
            let _ = push_proto sym in
            Ksi.Ref(extern_name sym)
          else
            const_uint (union_field_id sym) in
        let cond1 = Ksi.Binary(Ksi.Eq, ksi_un_id, ksi_f_id) in
        let (c', cond2) = match_it (c, ksi_val_ptr, val_pat) in
        (c', Ksi.Binary(Ksi.TruthAndIf, cond1, cond2))
    | UnionPattern {upr_sym = None} -> assert false 

and strip_patt_var patter =
  match patter.raw_pattern with
  | VoidPattern -> []
  | VarPattern {vp_symbol = Some sym} ->
      let loc = patter.pat_loc in
      if (not (has_property sym IsInPattern)) then
        if has_property sym IsInClosure then
          err(loc, "cannot use pattern variables in local functions")
        else
          (set_property sym IsInPattern;
           [sym])
      else err(loc, "multiple pattern variable occurrence")
  | VarPattern {vp_symbol = None} -> assert false
  | ExprPattern _ -> []
  | TuplePattern {tpr_patterns = tupl} ->
      List.flatten (List.map strip_patt_var tupl)   
  | UnionPattern {upr_pattern = pat} ->
      strip_patt_var pat

and match_let_patterns (c, pel, nr) =
   match pel with 
   | (pat, ex)::xs ->
      let ex_var = seq_name ("let_" ^ string_of_int(nr) ^ "_ex$") in
      let ex_ptr = Ksi.Ref(ksi_name ex_var) in
      let allocate_patt_var sym =
        let ksi_type_of_var = Ksi.Ptr(ksi_local_type sym.sym_type) in
        let value = Ksi.Convert(ksi_type_of_var, 
                                Ksi.Unary(Ksi.Addr, 
                                find_in_pattern sym.sym_name
                                  pat.raw_pattern 
                                  ex_ptr)) in
        Ksi.Compound(Ksi.Var(ksi_type_of_var, extern_name sym),
                     Ksi.Modify(Ksi.Ref(extern_name sym), value)) in
      let (c, exp) = ksi_exp (c, ex) in
      let varlist = strip_patt_var pat in
      let assigned_vars =
        Ksi.CompoundList(List.map allocate_patt_var varlist) in      
      let s = (Ksi.CompoundList [
         Ksi.Var (ksi_local_type (type_of ex), ksi_name ex_var);
         Ksi.Modify (ex_ptr, exp); assigned_vars]) in
      let (c, ss) = match_let_patterns(c, xs, nr+1) in
      (c, s::ss)
   | [] -> (c, []) 

and ksi_exp (c, e) =
  let rec ksi_call_parms (c, pt, p) =
    (* We need to create temps for all arguments, as order of evaluation
     * thereof is undefined in Ksi. *)
    let rec mktemps no r c p =
      match p with
          x::xs ->
            let t = instance_value (type_of x) in
            let (c', e) = ksi_exp (c, x) in
            let tmp_name = ksi_name ("_" ^ string_of_int(no) ^ "$") in
            let v = Ksi.Var(ksi_local_type (type_of x), tmp_name) in
            let vi = Ksi.Modify(Ksi.Ref(tmp_name), e) in
            mktemps (no+1) (vi::v::r) c' xs
        | [] -> (c, List.rev r) in
    (* Then compute parameter list. *)
    let rec mkexps no r c p pt =
      let convert t e =
        let t = instance_value t in
        match t with
            TyVar _ | AbstractType | Void -> 
              Ksi.Convert(Ksi.Ptr(Ksi.Void), e)
          | Int | Void | Float | String
          | Tuple _ | Func _ -> e
          | Bool | Struct _ | OptStruct _ | Union _ | Exn ->
              (* Fixup types of structs/unions (so they match after
               * conversion to ksi namespace) *)
              Ksi.Convert(ksi_passed_type t, e)
          | NamedType _ -> assert false in
      let (pt_hd, pt_tl) =
        match pt with
            x::xs -> (x, xs)
          | [] -> (TyVar "a", []) (* XXX I guess this was intended for
                                   * varargs...  --mal *) in
      match p with
        | x::xs ->
            let tmp_name = ksi_name ("_" ^ string_of_int(no) ^ "$") in
            let v = Ksi.Ref(tmp_name) in
            let e = convert pt_hd v in
            let e =
               match instance_value (type_of x) with
               | Float -> Ksi.Unary(Ksi.Addr, e)
               | _ -> e in
            mkexps (no+1) (e::r) c xs pt_tl
        | [] -> (c, List.rev r) in
    let (c', tmps) = mktemps 0 [] c p in
    let (c'', parms) = mkexps 0 [] c' p pt in
    (c'', tmps, parms) in
  let convert_return_value ret_type needed_type e =
    match instance_value ret_type with
      | TyVar _ ->
          let t = instance_value needed_type in
          (match t with
            | Float -> 
                Ksi.Indirect(Ksi.Convert(ksi_passed_type t, e))
            | _ -> 
                (match ksi_local_type needed_type with
                     Ksi.Ptr(Ksi.Void) -> e
                   | _ -> Ksi.Convert(ksi_passed_type needed_type, e)))
      | Float -> Ksi.Indirect(e)
      | Bool | Struct _ | OptStruct _ | Union _ ->
            Ksi.Convert(ksi_passed_type (instance_value needed_type), e)
      | _ -> e in
  let toplev_func sym =
    match instance_value sym.sym_type with
        Func _ when (sym.sym_scope_level <= 0 && 
                     not (has_property sym IsGlobalVar) &&
                     not (has_property sym IsUnionField)) -> true
      | _ -> false in
  let saved_e = e in
  let (c', e') =
    match e.raw_exp with
      | SimpleBinOp(Assign, e1, e2) ->
          begin
            let (c, exp) = ksi_exp (c, e2) in
            match e1.raw_exp with
            | Ref _
            | DotExp _ -> simple_assign c e1 exp
            | _ -> 
              let ex_var = seq_name "assign_ex$" in
              let ex_ptr = Ksi.Ref(ksi_name ex_var) in
              let (c, ass) = assign_proc (c, e1, ex_ptr) in
              (c, 
                Ksi.Bind (Ksi.CompoundList [
                   Ksi.Var (ksi_local_type (type_of e2), ksi_name ex_var);
                   Ksi.Modify (ex_ptr, exp); ass]))
          end
      | SimpleBinOp(op, e1, e2) -> 
          let (c1, e1') = ksi_exp (c, e1) in
          let (c2, e2') = ksi_exp (c1, e2) in
          let e' = 
            match (op, instance_value (type_of e1)) with
            (* String concatenation. *)
            | (Plus, String) -> Ksi.Call(ksi_name "gont_strjoin",
                                         [e1'; e2'])
            (* String comparison. *)
            (* We special case Eq and Ne, since, they are most often used,
             * and gont_streq is somewhat faster then gont_strcmp. *)
            | (Eq, String) -> 
                 Ksi.Convert(Ksi.Bool, Ksi.Call(ksi_name "gont_streq",
                                                [e1'; e2']))
            | (Ne, String) -> 
                 Ksi.Binary(Ksi.Ne, Ksi.Call(ksi_name "gont_streq",
                                             [e1'; e2']),
                                    const_int 0)
            | (Lt, String)
            | (Gt, String)
            | (Ge, String)
            | (Le, String) ->
                 Ksi.Binary(ksi_op2 op, Ksi.Call(ksi_name "gont_strcmp",
                                                 [e1'; e2']),
                                        const_int 0)
            (* Special case, we can't tell whatever to use TruncDiv or
             * RDiv in ksi_op2, without knowing the type. *)
            | (Div, Float) -> Ksi.Binary(Ksi.RDiv, e1', e2')
            | _ -> Ksi.Binary(ksi_op2 op, e1', e2') in
          (c2, e')
      | SimpleUnOp(op, e) -> 
          let (c', e') = ksi_exp (c, e) in
          (c', ksi_op1 op e')
      | Ref {ref_symbol = Some sym}
      | CrossModuleRef {xmr_symbol = Some sym}
                when (toplev_func sym) ->
        (* Need to encapsulate top level function into a function with 
         * closure. *)
        let _ = push_proto sym in
        let aux_name = sym.sym_name ^ "_" ^ (seq_name "closurized") ^ "$" in
        let (ret_type, parm_types) = 
          match instance_value sym.sym_type with
              Func {ft_ret_type=r; ft_parms=p} -> (r, p)
            | _ -> assert false in
        let cnt = ref 0 in
        let parm_names = 
          (List.map (fun _ -> incr cnt; 
                              ("_" ^ string_of_int(!cnt))) parm_types) in
        let call = Ksi.Call(extern_name sym, 
                            (List.map (fun n -> Ksi.Ref(ksi_name n)) 
                                      parm_names)) in
        let body = 
          (* Don't try to return void value from function (Ksi ICEs on 
           * that ;). *)
          match ret_type with
          | Void -> call
          | _ -> Ksi.Return(call) in
        let aux = 
           Ksi.Func([Ksi.Local], ksi_passed_type ret_type, ksi_name aux_name, 
                    (Ksi.Ptr(Ksi.Void), ksi_name "dummy_cl$") :: 
                      (List.map2 (fun t n -> (ksi_passed_type t, ksi_name n)) 
                                 parm_types parm_names), body) in
        let lambda_val = 
           Ksi.Call(ksi_name "gont_make_lambda", 
                    [Ksi.Convert(Ksi.Ptr(Ksi.Void), 
                                 Ksi.FuncRef(ksi_name aux_name));
                     ksi_null]) in
        ({c with ctx_aux_decls = aux :: c.ctx_aux_decls}, lambda_val)
      | CrossModuleRef {xmr_symbol = Some sym} ->
          let _ = push_proto sym in
          (c, Ksi.Ref(extern_name sym))
      | CrossModuleRef _ -> assert false
      | Ref ({ref_symbol = Some sym} as r) ->
          let _ = push_proto sym in
          let code = 
            let rec refer lev =
              if lev == 1 then 
                 ksi_parent_closure
              else 
                 Ksi.Arrow(refer (lev-1), ksi_name "parent_cl$") in
            if r.ref_level == 0 then
              if has_property sym IsInClosure then
                ksi_closure_var sym.sym_name
              else
                if has_property sym IsInPattern then
                  Ksi.Indirect(Ksi.Ref(extern_name sym))
                else
                  Ksi.Ref(ksi_name sym.sym_name)
            else
              (assert (has_property sym IsInClosure);
               Ksi.Arrow(refer r.ref_level, extern_name sym)) in
          let code = 
            match sym.sym_type with
            | Float when (has_property sym IsFormalParm) -> Ksi.Indirect(code)
            | Bool -> Ksi.Convert(Ksi.Bool, code)
            | _ -> code in
          (c, code)
      | Ref _ -> assert false
      (* The easy case -- global function call. *)
      | Call({raw_exp = Ref {ref_symbol = Some sym}}, p) 
        when (toplev_func sym) ->
          let _ = push_proto sym in
          let (ret_type, parm_types) = 
            match instance_value sym.sym_type with
                Func {ft_parms = x; ft_ret_type = r} -> (r, x)
              | _ -> assert false in
          let (c', t, p') = ksi_call_parms (c, parm_types, p) in
          let call = Ksi.Call(extern_name sym, p') in
          let call' = convert_return_value ret_type (type_of saved_e) call in
          let code = 
            match t with
                [] -> Ksi.Comment("global call, rettype = " ^ 
                                  (string_of_type (type_of saved_e)), call')
              | _ -> Ksi.Bind(Ksi.CompoundList(t @ [call'])) in
          (c', code)
      (* The hard case -- function call with closure. *)
      | Call(e, p) ->
          let (ret_type, parm_types) =
            match instance_value (type_of e) with
                Func {ft_ret_type=r; ft_parms=p} -> (r, p)
              | _ -> assert false in
          let (c', e') = ksi_exp (c, e) in
          let (c'', t, p') = ksi_call_parms (c', parm_types, p) in
          let lambda_var = 
             Ksi.Var(Ksi.Ptr(Ksi.Struct (ksi_name "gont_lambda_exp")), 
                     ksi_name "lambda$") in
          let lambda_init =
             Ksi.Modify(Ksi.Ref(ksi_name "lambda$"), e') in
          let lambda_fld n = 
             Ksi.Arrow(Ksi.Ref(ksi_name "lambda$"), ksi_name n) in
          let ksi_fnc_type = 
             Ksi.FuncPtr(ksi_passed_type ret_type,
                         Ksi.Ptr(Ksi.Void) :: 
                         (List.map ksi_passed_type parm_types)) in
          let fnc_ptr = Ksi.Convert(ksi_fnc_type, lambda_fld "fnc") in
          let call = Ksi.PtrCall(fnc_ptr, (lambda_fld "closure") :: p') in
          let call' = convert_return_value ret_type (type_of saved_e) call in
          (c'', Ksi.Bind(Ksi.CompoundList(lambda_var :: 
                                          lambda_init :: t @ [call'])))
      | IntConst(i) -> (c, Ksi.IntConst(ksi_local_type Int, i))
      (* We lay here... but this value is sometimes needed... *)
      | VoidConst -> (c, ksi_null)
      | BoolConst(x) -> (c, Ksi.BoolConst(x))
      | FloatConst(f) -> (c, Ksi.FloatConst(ksi_local_type Float, f))
      | StringConst(s) -> 
         (c, Ksi.Call(ksi_name "gont_make_string", 
                      [Ksi.IntConst(Ksi.uptroff(), 
                                    string_of_int(String.length s));
                       Ksi.StringConst(s)]))
      | NullConst -> (c, ksi_null)
      | TupleExp(tup) ->
         let (c, tmps, parms) = 
                ksi_call_parms (c, List.map type_of tup, tup) in
         let call = Ksi.Call(ksi_name "gont_make_tuple",
                             (const_uint (List.length parms)) :: parms) in
         (c, Ksi.Bind(Ksi.CompoundList(tmps @ [call])))
      | UnionExp {ur_sym = Some sym; ur_exp = unex} ->
          let (c', valu) = ksi_exp (c, unex) in
          let valu = Ksi.Convert(Ksi.Ptr(Ksi.Void), valu) in
          let un_id =
            if has_property sym IsException then
              let _ = push_proto sym in
              Ksi.Ref(extern_name sym)
            else
              const_uint (union_field_id sym) in
          (c', Ksi.Convert(ksi_local_type (type_of e),
                           Ksi.Call(ksi_name "gont_make_union",
                                    [un_id; valu])))
      | UnionExp {ur_sym = None} -> assert false
      | StructExp sis -> 
         let ptr_name = seq_name "sinit$" in
         let init_one (c, si) =
           match si with
           | {si_expr = e; si_sym = Some sym} ->
             let lval = 
               Ksi.Arrow(Ksi.Ref(ksi_name ptr_name), extern_name sym) in
             let (c, rval) = ksi_exp (c, e) in
             (c, Ksi.Modify(lval, rval)) 
           | {si_sym = None} -> assert false in
         let var_def = 
            Ksi.Var(ksi_local_type (type_of e), 
                    ksi_name ptr_name) in
         let init = 
            Ksi.Modify(Ksi.Ref(ksi_name ptr_name), 
                       Ksi.Convert(ksi_local_type (type_of e),
                         Ksi.Call(ksi_name "gont_xmalloc",
                                  [Ksi.SizeOf(ksi_noptr_type (type_of e))]))) in
         let (c, inits) = list_apply init_one (c, sis) in
         (c, Ksi.Compound (Ksi.CompoundList (var_def :: init :: inits), 
                           Ksi.Ref(ksi_name ptr_name)))
      | DotExp {de_expr = ex; de_sym = Some sym} ->
          let (c, e) = ksi_exp (c, ex) in
          let e = 
            match instance_value (type_of ex) with
            | OptStruct _ ->
              let n = ksi_name "nullcheck$" in
                Ksi.Convert(ksi_local_type (type_of ex),
                  Ksi.Bind(
                    Ksi.CompoundList [
                      Ksi.Var(ksi_local_type (type_of ex), n);
                      Ksi.Modify(Ksi.Ref(n), e);
                      Ksi.Cond(Ksi.Binary(Ksi.Eq, Ksi.Ref(n), ksi_null),
                               Ksi.Call(ksi_name "gont_null_access", []),
                               Ksi.Ref(n))
                    ]))
            | Struct _ -> e
            | _ -> assert false in
          (c, Ksi.Arrow(e, extern_name sym))
      | DotExp {de_sym = None} ->
          assert false
      | Lambda {lambda_parms=p; lambda_ret_type=t; lambda_body=s} ->
        let aux_name = (seq_name "lambda") ^ "$" in
        let aux_decls = ksi_fnc_decl [Ksi.Local] c (t, p) aux_name s in
        let lambda_val = 
          Ksi.Call(ksi_name "gont_make_lambda", 
                   [Ksi.Convert(Ksi.Ptr(Ksi.Void), 
                                Ksi.FuncRef(ksi_name aux_name));
                    if c.ctx_closure_type = "" then
                      ksi_null
                    else
                      ksi_closure]) in
        ({c with ctx_aux_decls = aux_decls @ c.ctx_aux_decls}, lambda_val) in
  (c', Ksi.Line(e.exp_loc.loc_line, e'))

and simple_assign c e1 e2 =
  match e1.raw_exp with
  | Ref {ref_symbol = Some sym} ->
      if has_property sym IsNonLvalue then
        err (e1.exp_loc, xf "changing a whole matched expression \
                             is not allowed")
      else
        let (c, e1') = ksi_exp (c, e1) in
        (c, Ksi.Modify(e1', Ksi.Convert(ksi_local_type (type_of e1), e2)))
  | DotExp _ ->
     let (c, e1') = ksi_exp (c, e1) in
     (c, Ksi.Modify(e1', e2))
  | _ -> assert false 
   
and assign_proc (c, ex, e_ptr) = 
  match ex.raw_exp with
    | Ref _ 
    | DotExp _ ->
        simple_assign c ex e_ptr
    | TupleExp tl ->
      let rec assign_list (c, idx, lst) =
        match lst with
        | x::xs ->
          (* *( ((void ** )ksi_ptr) + (idx * sizeof(void* )) ) *)
          let n_ptr = (Ksi.Indirect(Ksi.Binary(Ksi.Plus, 
                        Ksi.Convert(Ksi.Ptr(Ksi.Ptr(Ksi.Void)), e_ptr), 
                        Ksi.Binary(Ksi.Mult, 
                                   const_uint idx, 
                                   Ksi.SizeOf(Ksi.Ptr(Ksi.Void)))))) in
          let (c', x') = assign_proc (c, x, n_ptr) in
          let (c'', xs') = assign_list (c', idx+1, xs) in
          (c'', Ksi.Compound(x',xs')) 
        | [] -> (c, Ksi.Nop) in
      assign_list (c, 0, tl)                         
    | _ -> assert false
  
and ksi_fnc_decl storage c1 (ret_type, parms) fnname body =
  let closure_struct name cl =
    let flds = (List.map (fun sym -> (ksi_passed_type sym.sym_type, 
                                      extern_name sym)) cl) in
    let flds = (Ksi.uptroff(), ksi_name "dummy$") :: flds in
    match c1.ctx_closure_type with
        "" -> Ksi.StructDef(name, [], flds)
      | _ -> let parent_cl = (Ksi.Ptr(Ksi.Struct(ksi_name c1.ctx_closure_type)), 
                              ksi_name "parent_cl$") in
             Ksi.StructDef(name, [], parent_cl :: flds) in
  let (need_cl, cl) = closure (c1.ctx_level + 1) body in
  let c2 = {ctx_level = c1.ctx_level + 1;
            ctx_closure_type = "";
            ctx_aux_decls = [];
            ctx_fnname = fnname;
            ctx_module = c1.ctx_module} in
  let c3 = if not need_cl then c2 else
           let name = fnname ^ "_" ^ (seq_name "closure") ^ "_t$" in
               {c2 with
                ctx_closure_type = name;
                ctx_aux_decls = (closure_struct (ksi_name name) cl) :: 
                                c2.ctx_aux_decls} in
  let (c4, code) = ksi_stmt (c3, body) in
  let code_with_cl =
    if not need_cl then 
      code 
    else
      let cl_struct = Ksi.Struct(ksi_name c3.ctx_closure_type) in
      let cl_var = Ksi.Var(Ksi.Ptr(cl_struct), ksi_name "cl$") in
      let cl_init = 
          Ksi.Modify(ksi_closure,
                     Ksi.Convert(Ksi.Ptr(cl_struct),
                                 Ksi.Call(ksi_name "gont_xmalloc0",
                                          [Ksi.SizeOf(cl_struct)]))) in
      let cl_copy_parent =
        match c1.ctx_closure_type with
            "" -> Ksi.Nop
          | _ -> Ksi.Modify(ksi_closure_var "parent_cl$",
                            ksi_parent_closure) in
      let cl_copy_parms =
         let rec helper parms =
           match parms with
               {parm_symbol = Some sym}::xs ->
                 let rest = helper xs in
                 if has_property sym IsInClosure then
                   Ksi.Modify(ksi_closure_var sym.sym_name, 
                              Ksi.Ref(extern_name sym)) :: rest
                 else
                    rest
             | [] -> []
             | _ -> assert false in
          helper parms in
      Ksi.CompoundList([cl_var; cl_init; cl_copy_parent] @ 
                       cl_copy_parms @ 
                       [code]) in
  let (ksi_ret_type, ksi_parms) = 
         (ksi_passed_type ret_type, 
          (List.map (fun {parm_type=t; parm_name=n} -> 
                                  (ksi_passed_type t, ksi_name n))
                    parms)) in
  let ksi_parms' =
    if c1.ctx_closure_type <> "" then
      (Ksi.Ptr(Ksi.Struct(ksi_name c1.ctx_closure_type)), 
       ksi_name "parent_cl$") 
      :: ksi_parms
    else if c1.ctx_level > 0 then
      (Ksi.Ptr(Ksi.Void), ksi_name "dummy_cl$") :: ksi_parms
    else
      ksi_parms in
  let func_decl = 
    Ksi.Func(storage,
             ksi_ret_type,
             ksi_name fnname,
             ksi_parms',
             code_with_cl) in
  match c4.ctx_aux_decls with
      [] -> 
        [Ksi.TopComment("END " ^ fnname); 
         func_decl; 
         Ksi.TopComment("BEG " ^ fnname)]
    | aux -> 
        Ksi.TopComment("END " ^ fnname) ::
        func_decl ::
        Ksi.TopComment("BEG " ^ fnname) :: 
        aux @ 
        [Ksi.TopComment("AUX " ^ fnname)]
;;

let wrap_as name sym =
  let (ret_type, parm_types) = 
    match instance_value sym.sym_type with
        Func {ft_ret_type=r; ft_parms=p} -> (r, p)
      | _ -> assert false in
  let cnt = ref 0 in
  let parm_names = 
    (List.map (fun _ -> incr cnt; 
                        ("_" ^ string_of_int(!cnt))) parm_types) in
   Ksi.Func([Ksi.Public], ksi_passed_type ret_type, ksi_name name, 
            (List.map2 (fun t n -> (ksi_passed_type t, ksi_name n)) 
                       parm_types parm_names),
            Ksi.Return(Ksi.Call(extern_name sym, 
              (List.map (fun n -> Ksi.Ref(ksi_name n)) parm_names)))) 
;;

let ksi_code ctx d =
  let (ctx', decls) = 
    match d with
        FuncDecl ({fn_symbol = Some sym; fn_loc = l} as f) ->
          let c = {global_ctx with ctx_module = ctx.gctx_module} in
          let (ctx', storage) = global_check l ctx sym in
          let (ctx', storage) = 
            match f.fn_name with
            | "init$" ->
              ({ctx' with gctx_init_fnc = "init$"}, [Ksi.Public])
            | "fini$" ->
              ({ctx' with gctx_fini_fnc = "fini$"}, [Ksi.Public])
            | _ -> (ctx', storage) in
          let decls = (ksi_fnc_decl storage c
                                    (f.fn_ret_type, f.fn_parms)
                                    sym.sym_name 
                                    f.fn_body) in
          let proto = Ksi.FuncProtoNamed(storage, 
                                    ksi_passed_type f.fn_ret_type,
                                    extern_name sym,
                                    (List.map 
                                        (fun {parm_type = t; parm_name = n} -> 
                                                (ksi_passed_type t,
                                                 ksi_name n))
                                        f.fn_parms)) in
          if f.fn_export_as = "" then
            (ctx', proto :: (List.rev decls))
          else
            (ctx', proto :: (List.rev (wrap_as f.fn_export_as sym :: decls)))
      | FuncDecl _ -> assert false
      | GlobalDecl {global_symbol=Some sym; global_loc = l} ->
           let (ctx', storage) = (global_check l ctx sym) in
           (ctx', [Ksi.Global(storage,
                              ksi_local_type sym.sym_type,
                              extern_name sym)])
      | GlobalDecl _ -> assert false
      | ExternC {xc_symbol = Some sym} ->
         (ctx,
           (match instance_value sym.sym_type with
               Func {ft_ret_type=r; ft_parms=a} ->
                 [Ksi.FuncProto([Ksi.Extern], ksi_passed_type r,
                                extern_name sym,
                                (List.map ksi_passed_type a))]
             | t ->
                 [Ksi.Global([Ksi.Extern], ksi_passed_type t, 
                             extern_name sym)]))
      | ExternC _ -> assert false
      | FuncProto _ -> assert false
      | TopLevelOpen _
      | ExceptionDecl _
      | SectionDecl _
      | DummyDecl
      | TypeDef _ -> (ctx, []) in
   let decls' = (List.rev !aux_decls) @ decls in
   aux_decls := [];
   (ctx', decls')
;;

let check_module ctx =
  match (List.filter (fun s -> not (has_property s IsUnionField) &&
                               not (has_property s IsException) &&
                               not (List.exists 
                                      (function 
                                       | ExternName _ -> true 
                                       | _ -> false) s.sym_properties) &&
                               (s.sym_kind <> StructFieldKind))
                     ctx.gctx_globals) with
      [] -> ()
    | l -> 
      let oops = String.concat ", " (List.map (fun sym -> sym.sym_name) l) in
      raise (Error ("following symbols, required by interface, are not \
                     provided: " ^ oops))
;;
